package com.sillycycle.bagleyd.mball;

/*
 * @(#)MballCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>MballCanvas</code> class does all the graphics for
 * <code>MballFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class MballCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 64;

	static final int TOP = 0;
	static final int TTR = 1;
	static final int TR = 2;
	static final int RIGHT = 3;
	static final int BR = 4;
	static final int BBR = 5;
	static final int BOTTOM = 6;
	static final int BBL = 7;
	static final int BL = 8;
	static final int LEFT = 9;
	static final int TL = 10;
	static final int TTL = 11;
	static final int COORD = 12;
	static final int CW = 13;
	static final int CCW = 23;
	static final int CUTS = (COORD >> 1);
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int MAX_VIEWS = 2;
	static final int NUM_DEGREES = 360;
	static final int ST_ANG = (NUM_DEGREES >> 1);
	static final int RT_ANG = (NUM_DEGREES >> 2);

	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "mball.dat";
	static final String SYMBOL = ":";

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     wedgeColor = new Color[MballInterface.MAX_WEDGES];
	int	 currentDirection, lastDirection;
	boolean     nextShift = false;
	boolean     currentControl = false, nextControl = false;
	int	 currentSector = -1, currentView = -1;
	int	 nextSector = -1, nextView = -1;
	int	 rowPosition;
	boolean     started, cheat, vertical, perspective;
	boolean     mono, reverse, orient, practice;
	int	 base;
	int	 delay;
	int	 lastX, lastY, dr;
	int	 bands, wedges, wedgeBands;
	Point       coreSize;
	Point       offset, letter;
	Point       puzzleSize, puzzleOffset, offsetD;
	int	 delta, mballLength, wedgeLength, viewLength;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allWedgesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	Random      generator;
	transient MballStack  undo = null, redo = null;
	transient MballSolve  solve = null;
	String      stringSave = null;
	String[]    token = null;
	String      wedgeNames = null;

	transient MballLoc[]  wedgeLoc = null;
	transient MballLoc[]  startLoc = null;

static final int[][] /*[(MballInterface.MAX_WEDGES - MballInterface.MIN_WEDGES) / 2 + 1][COORD]*/ mapDirToWedge =
{
	{
		0, 6, 6, 6, 6, 6, 0, 6, 6, 6, 6, 6
	},
	{
		0, 6, 6, 1, 6, 6, 0, 6, 6, 1, 6, 6
	},
	{
		0, 6, 1, 6, 2, 6, 0, 6, 1, 6, 2, 6
	},
	{
		0, 6, 1, 2, 3, 6, 0, 6, 1, 2, 3, 6
	},
	{
		0, 1, 2, 6, 3, 4, 0, 1, 2, 6, 3, 4
	},
	{
		0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5
	}
};

static final int[][] /*[(MballInterface.MAX_WEDGES - MballInterface.MIN_WEDGES) / 2 + 1][MballInterface.MAX_WEDGES]*/ mapWedgeToDir =
{
	{
		0,  6, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12
	},
	{
		0,  3,  6,  9, 12, 12, 12, 12, 12, 12, 12, 12
	},
	{
		0,  2,  4,  6,  8, 10, 12, 12, 12, 12, 12, 12
	},
	{
		0,  2,  3,  4,  6,  8,  9, 10, 12, 12, 12, 12
	},
	{
		0,  1,  2,  4,  5,  6,  7,  8, 10, 11, 12, 12
	},
	{
		0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11
	}
};

	static double radians(double x) {
		return x * 2.0 * Math.PI / NUM_DEGREES;
	}

	@Override
	protected String paramString() {
		return super.paramString() +
		",orient=" + orient +
		",practice=" + practice +
		",bands=" + bands +
		",wedges=" + wedges +
		",base=" + base +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + wedgeNames +
		",color0=" + wedgeColor[0] +
		",color1=" + wedgeColor[1] +
		",color2=" + wedgeColor[2] +
		",color3=" + wedgeColor[3] +
		",color4=" + wedgeColor[4] +
		",color5=" + wedgeColor[5] +
		",color6=" + wedgeColor[6] +
		",color7=" + wedgeColor[7] +
		",color8=" + wedgeColor[8] +
		",color9=" + wedgeColor[9] +
		",color10=" + wedgeColor[10] +
		",color11=" + wedgeColor[11] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay +
		",perspective=" + perspective;
	}

	void checkPieces() {
		if (bands < MballInterface.MIN_BANDS) {
			System.out.println(bands +
			" number of mball in a row incorrect, must be at least " +
			MballInterface.MIN_BANDS +
			", defaulting to " +
			MballInterface.DEFAULT_BANDS);
			bands = MballInterface.DEFAULT_BANDS;
		}
		if (wedges < MballInterface.MIN_WEDGES ||
				wedges > MballInterface.MAX_WEDGES ||
				(wedges & 1) == 1) {
			System.out.println(wedges +
			" number of mball in a row incorrect, must be at between " +
			MballInterface.MIN_WEDGES + " and " +
			MballInterface.MAX_WEDGES +
			"inclusive and even, defaulting to " +
			MballInterface.DEFAULT_WEDGES);
			bands = MballInterface.DEFAULT_WEDGES;
		}
		if (base < MballInterface.MIN_BASE ||
				base > MballInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			MballInterface.MIN_BASE +
			" and " +
			MballInterface.MAX_BASE +
			" inclusive, defaulting to " +
			MballInterface.DEFAULT_BASE);
			base = MballInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	int toWedge(int pos) {
		return (pos % bands);
	}

	int toBand(int pos) {
		return (pos / bands);
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		MballLoc tempLoc = new MballLoc(wedgeLoc[0].wedge, wedgeLoc[0].direction);

		if (orient) {
		  for (int wedge = 0; wedge < wedges; wedge++) {
		    for (int band = 0; band < bands; band++) {
		      if (tempLoc.direction != wedgeLoc[band * wedges + wedge].direction)
			return false;
		      if (tempLoc.direction == 0) {
			if ((wedges + wedgeLoc[band * wedges + wedge].wedge -
			    tempLoc.wedge) % wedges != wedge)
			  return false;
		      } else {
			if ((wedges - wedgeLoc[band * wedges + wedge].wedge +
			    tempLoc.wedge) % wedges != wedge)
			  return false;
		      }
		    }
		  }
		} else {
		  for (int wedge = 0; wedge < wedges; wedge++) {
		    tempLoc.wedge = wedgeLoc[wedge].wedge;
		    for (int band = 1; band < bands; band++)
		      if (tempLoc.wedge != wedgeLoc[band * wedges + wedge].wedge)
			return false;
		  }
		}
		started = false;
		return true;
	}

	void letterPosition(int wedge, int band, int lengthx, int lengthy) {
		double angle, radius;

		angle = (2 * wedge + 1) * Math.PI / wedges;
		if (perspective) {
			if (((bands & 1) == 1) && band == bands / 2)
				radius = Math.sin(Math.PI * (0.25 + band) / bands) / 2.0;
			else
				radius = Math.sin(Math.PI * (0.5 + band) / bands) / 2.0;
		} else {
			if (((bands & 1) == 1) && band == bands / 2)
				radius = (4.0 * band + 1.0) / (4.0 * bands);
			else
				radius = (2.0 * band + 1.0) / (2.0 * bands);
		}
		letter = new Point(lengthx / 2 + (int) (lengthx *
			radius * Math.cos(angle - Math.PI / 2.0)),
			lengthy / 2 + (int) (lengthy * radius *
			Math.sin(angle - Math.PI / 2.0)));
	}

	void offsetSect(int wedge) {
		double angle = (2 * wedge + 1) * Math.PI / wedges;

		offsetD = new Point((int) (dr * Math.cos(angle - Math.PI / 2)),
			(int) (dr * Math.sin(angle - Math.PI / 2)));
	}

	void drawSect(Graphics g, Color pieceColor, Color border,
			int r, int wedge, int startx, int starty,
			int lengthx, int lengthy) {
		int dx, dy, diam = 2 * dr;
		double u, v;

		offsetSect(wedge);
		dx = offsetD.x;
		dy = offsetD.y;
		if ((bands & 1) == 1) {
			if (r == bands / 2) {
				if (perspective) {
					u = Math.sin(Math.PI * r / bands) / 2.0;
					v = 0.5;
				} else {
					u = (double) r / bands;
					v = (double) (r + 1) / (bands + 1);
				}
				g.setColor(pieceColor);
				fillSector(g,
					startx + lengthx / 2 + dx,
					starty + lengthy / 2 + dy,
					(int) (u * 2 * lengthx - diam),
					(int) (u * 2 * lengthy - diam),
					(int) (v * 2 * lengthx - diam),
					(int) (v * 2 * lengthy - diam),
					RT_ANG - NUM_DEGREES * wedge / wedges,
					-NUM_DEGREES / wedges);
				g.setColor(border);
				drawSector(g,
					startx + lengthx / 2 + dx,
					starty + lengthy / 2 + dy,
					(int) (u * 2 * lengthx - diam),
					(int) (u * 2 * lengthy - diam),
					(int) (v * 2 * lengthx - diam),
					(int) (v * 2 * lengthy - diam),
					RT_ANG - NUM_DEGREES * wedge / wedges,
					-NUM_DEGREES / wedges);
			} else {
				if (perspective) {
					u = Math.sin(Math.PI * r / bands) / 2.0;
					v = Math.sin(Math.PI * (r + 1) / bands) / 2.0;
				} else {
					u = (double) r / bands;
					v = (double) (r + 1) / bands;
				}
				g.setColor(pieceColor);
				fillSector(g,
					startx + lengthx / 2 + dx,
					starty + lengthy / 2 + dy,
					(int) (u * 2 * lengthx - diam),
					(int) (u * 2 * lengthy - diam),
					(int) (v * 2 * lengthx - diam),
					(int) (v * 2 * lengthy - diam),
					RT_ANG - NUM_DEGREES * wedge / wedges,
					-NUM_DEGREES / wedges);
				g.setColor(border);
				drawSector(g,
					startx + lengthx / 2 + dx,
					starty + lengthy / 2 + dy,
					(int) (u * 2 * lengthx - diam),
					(int) (u * 2 * lengthy - diam),
					(int) (v * 2 * lengthx - diam),
					(int) (v * 2 * lengthy - diam),
					RT_ANG - NUM_DEGREES * wedge / wedges,
					-NUM_DEGREES / wedges);
			}
		} else {
			if (perspective) {
				u = Math.sin(Math.PI * r / bands) / 2.0;
				v = Math.sin(Math.PI * (r + 1) / bands) / 2.0;
			} else {
				u = (double) r / bands;
				v = (double) (r + 1) / bands;
			}
			g.setColor(pieceColor);
			fillSector(g,
				startx + lengthx / 2 + dx,
				starty + lengthy / 2 + dy,
				(int) (u * 2 * lengthx - diam),
				(int) (u * 2 * lengthy - diam),
				(int) (v * 2 * lengthx - diam),
				(int) (v * 2 * lengthy - diam),
				RT_ANG - NUM_DEGREES * wedge / wedges,
				-NUM_DEGREES / wedges);
			g.setColor(border);
			drawSector(g,
				startx + lengthx / 2 + dx,
				starty + lengthy / 2 + dy,
				(int) (u * 2 * lengthx - diam),
				(int) (u * 2 * lengthy - diam),
				(int) (v * 2 * lengthx - diam),
				(int) (v * 2 * lengthy - diam),
				RT_ANG - NUM_DEGREES * wedge / wedges,
				-NUM_DEGREES / wedges);
		}
	}

	static void drawSector(Graphics g, int xo, int yo,
			int width1, int height1, int width2, int height2,
			int angle1, int angle2) {
		int d;
		int r1 = java.lang.Math.min(width1, height1) / 2;
		int r2 = java.lang.Math.min(width2, height2) / 2;

		if (r1 > r2) {
			d = r1;
			r1 = r2;
			r2 = d;
		}
		if (r1 < 0)
			r1 = -3;
		d = java.lang.Math.max(2 * (r1 + 3), 1);
		g.drawArc(xo - d / 2, yo - d / 2, d, d,
			angle1, angle2);
		d = java.lang.Math.max(2 * (r2 - 1), 3);
		g.drawArc(xo - d / 2, yo - d / 2, d, d,
			angle1, angle2);

		/*
		ang = RADIANS(angle1 / MULT);
		x = Math.cos(ang);
		y = Math.sin(ang);
		g.drawLine((int) (r1 * x) + xo,
			(int) (r1 * y) + yo,
			(int) (r2 * x) + xo,
			(int) (r2 * y) + yo);
		ang = RADIANS(angle2 / MULT);
		x = Math.cos(ang);
		y = Math.sin(ang);
		g.drawLine((int) (r1 * x) + xo,
			(int) (r1 * y) + yo,
			(int) (r2 * x) + xo,
			(int) (r2 * y) + yo);
		*/
	}

	static void fillSector(Graphics g, int xo, int yo,
			int width1, int height1, int width2, int height2,
			int angle1, int angle2) {
		int d;
		int r1 = java.lang.Math.min(width1, height1) / 2;
		int r2 = java.lang.Math.min(width2, height2) / 2;

		if (r1 > r2) {
			d = r1;
			r1 = r2;
			r2 = d;
		}
		if (r1 < 0)
			r1 = -3;
		for (d = 2 * (r1 + 3); d < 2 * (r2 - 1); d++)
			g.drawArc(xo - d / 2, yo - d / 2, d, d,
				angle1, angle2);
	}

	/* If only you could set line thickness in Java */
	/* void fillSector(Graphics g, int xo, int yo,
			int width1, int height1, int width2, int height2,
			int angle1, int angle2) {
		int d;
		int r1 = java.lang.Math.min(width1, height1) / 2;
		int r2 = java.lang.Math.min(width2, height2) / 2;
		int w = java.lang.Math.max(r2 - r1 - 8, 1);

		if (r1 > r2) {
			d = r1;
			r1 = r2;
			r2 = d;
		}
		if (r1 < 0)
			r1 = -3;
		d = java.lang.Math.max(r1 + r2 + 2, 2);
		setLineAttributes(w, LineSolid, CapNotLast, JoinRound);
		g.drawArc(xo - d / 2, yo - d / 2, d, d,
			angle1, angle2);
		setLineAttributes(1, LineSolid, CapNotLast, JoinRound);
	} */

	void drawSector(int position, int offsetPosition) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			Point start = new Point(1 + puzzleOffset.x,
				1 + puzzleOffset.y);
			Point length = new Point(viewLength - delta +
				puzzleOffset.x, viewLength - delta +
				puzzleOffset.y);
			int wedge = position % wedges;
			int band = position / wedges;
			int l;
			Color fillColor, lineColor;

			l = wedgeLoc[position].wedge;
			if (offsetPosition == 0) {
				if (mono) {
					if (reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = wedgeColor[l];
					lineColor = borderColor;
				}
			} else {
				if (mono) {
					if (reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = wedgeColor[l];
					fillColor = borderColor;
				}
			}
			/* if the number of bands is odd
			 * the band can straddle both hemispheres */
			if (band < (bands + 1) / 2)
				drawSect(g, fillColor, lineColor,
					band, wedge,
					start.x, start.y,
					length.x - start.x, length.y - start.y);
			if (band + 1 > bands / 2) {
				if (vertical)
				    drawSect(g, fillColor, lineColor,
					bands - 1 - band,
					(3 * wedges / 2 - 1 - wedge) % wedges,
					start.x, length.y + 3, length.x - start.x,
					length.y - start.y);
				else
				    drawSect(g, fillColor, lineColor,
					bands - 1 - band,
					wedges - 1 - wedge,
					length.x + 3, start.y, length.x - start.x,
					length.y - start.y);
			}
			l++;
			if (mono) {
				String buf;
				int fontSize, b = bands;

				if (bands == 1)
					b = 2;
				g.setColor(lineColor);
				fontSize = java.lang.Math.min(length.x,
					length.y) / (5 * b);
				g.setFont(new Font("arial", Font.PLAIN, fontSize));
				if (band < (bands + 1) / 2) {
				    letterPosition(wedge, band,
					length.x - start.x,
					length.y - start.y);
				    letter.x += start.x;
				    letter.y += start.y + 2 * fontSize / 5;
				    if (orient && wedgeLoc[position].direction == DOWN) {
					buf = Integer.toString(l, base).toUpperCase() +
						wedgeNames.charAt(l - 1);
					if (l == 0)
					    l = 1;
					l *= base;
					while (l >= 1) {
					    l /= base;
					    letter.x -= fontSize / 3;
					}
				    } else {
					buf = Character.toString(wedgeNames.charAt(l - 1));
				    }
				    g.drawString(buf, letter.x, letter.y);
				}
				if (band + 1 > bands / 2) {
				    if (vertical) {
					letterPosition((3 * wedges / 2 - 1 -
					    wedge) % wedges, bands - 1 - band,
					    length.x - start.x, length.y - start.y);
					letter.x += start.x;
					letter.y += length.y + 3 + 2 * fontSize / 5;
				    } else {
					letterPosition(wedges - 1 - wedge,
					    bands - 1 - band,
					    length.x - start.x, length.y - start.y);
					letter.x += length.x + 3;
					letter.y += start.y + 2 * fontSize / 5;
				    }
				    if (orient && wedgeLoc[position].direction == UP) {
					buf = Integer.toString(l, base).toUpperCase() +
						wedgeNames.charAt(l - 1);
					if (l == 0)
					    l = 1;
					l *= base;
					while (l >= 1) {
					    l /= base;
					    letter.x -= fontSize / 3;
					}
				    } else {
					buf = Character.toString(wedgeNames.charAt(l - 1));
				    }
				    g.drawString(buf, letter.x, letter.y);
				}
			} else if (orient) {
				String buf;
				int fontSize, b = bands;

				if (bands == 1)
					b = 2;
				g.setColor(lineColor);
				fontSize = java.lang.Math.min(length.x,
					length.y) / (5 * b);
				g.setFont(new Font("arial", 0, fontSize));
				if (band < (bands + 1) / 2 &&
					wedgeLoc[position].direction == DOWN) {
				    letterPosition(wedge, band,
					length.x - start.x, length.y - start.y);
				    letter.x += start.x;
				    letter.y += start.y + 2 * fontSize / 5;
				    buf = Integer.toString(l, base).toUpperCase();
				    if (l == 0)
					l = 1;
				    while (l >= 1) {
					l /= base;
					letter.x -= fontSize / 3;
				    }
				    g.drawString(buf, letter.x, letter.y);
				}
				if (band + 1 > bands / 2 &&
					wedgeLoc[position].direction == UP) {
				    if (vertical) {
					letterPosition((3 * wedges / 2 - 1 -
					    wedge) % wedges, bands - 1 - band,
					    length.x - start.x,
					    length.y - start.y);
					letter.x += start.x;
					letter.y += length.y + 3 + 2 * fontSize / 5;
				    } else {
					letterPosition(wedges - 1 - wedge,
					    bands - 1 - band,
					    length.x - start.x,
					    length.y - start.y);
					letter.x += length.x + 3;
					letter.y += start.y + 2 * fontSize / 5;
				    }
				    buf = Integer.toString(l, base).toUpperCase();
				    if (l == 0)
					l = 1;
				    while (l >= 1) {
					l /= base;
					letter.x -= fontSize / 3;
				    }
				    g.drawString(buf, letter.x, letter.y);
				}
			}
		} finally {
			g.dispose();
		}
	}

	void drawWedge(int wedge) {
		for (int band = 0; band < bands; band++)
			drawSector(wedge + wedges * band, 0);
	}

	void drawAllPieces() {
		for (int wedge = 0; wedge < wedges; wedge++)
			drawWedge(wedge);
	}

	void drawRadar(Graphics g,
			int startx, int starty, int lengthx, int lengthy) {
		int b, i, stx, sty;
		double angle, increment, u;

		g.drawArc(startx, starty,
			lengthx, lengthy, 0, NUM_DEGREES);
		if ((bands & 1) == 1) {
			stx = startx - lengthx / (2 * bands) +
				(bands / 2 + 1) * lengthx / bands;
			sty = starty - lengthy / (2 * bands) +
				(bands / 2 + 1) * lengthy / bands;
			for (b = 1; b < bands / 2 + 1; b++) {
				if (perspective) {
					u = Math.sin(Math.PI * b / bands) / 2.0;
				} else {
					u = (double) b / bands;
				}
				g.drawArc(
					(int) (stx - lengthx * u),
					(int) (sty -  lengthy * u),
					(int) (2 * lengthx * u),
					(int) (2 * lengthx * u),
					0, NUM_DEGREES);
			}
		} else {
			stx = startx + (bands / 2) * lengthx / bands;
			sty = starty + (bands / 2) * lengthy / bands;
			for (b = 1; b < bands / 2; b++) {
				if (perspective) {
					u = Math.sin(Math.PI * b / bands) / 2.0;
				} else {
					u = (double) b / bands;
				}
				g.drawArc(
					(int) (stx - lengthx * u),
					(int) (sty -  lengthy * u),
					(int) (2 * lengthx * u),
					(int) (2 * lengthx * u),
					0, NUM_DEGREES);
			}
		}
		increment = radians(NUM_DEGREES) / wedges;
		angle = radians(RT_ANG);
		for (i = 0; i < wedges; i++) {
			g.drawLine(
				startx + lengthx / 2, starty + lengthy / 2,
				startx + lengthx / 2 + (int) (lengthx *
					Math.cos(angle) / 2.0),
				starty + lengthy / 2 + (int) (lengthy *
					Math.sin(angle) / 2.0));
			angle += increment;
		}
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			Point start = new Point(1 + puzzleOffset.x,
				1 + puzzleOffset.y);
			Point length = new Point(viewLength - delta +
				puzzleOffset.x, viewLength - delta +
				puzzleOffset.y);
			if (show) {
				g.setColor(currentForeground);
				drawRadar(g,
					start.x, start.y, length.x - start.x,
					length.y - start.y);
				if (vertical) {
					g.setColor(foreground);
					g.drawLine(0, length.y + 1,
						getSize().width - 1,
						length.y + 1);
					g.setColor(currentForeground);
					drawRadar(g, start.x, length.y + 3,
						length.x - start.x,
						length.y - start.y);
				} else {
					g.setColor(foreground);
					g.drawLine(length.x + 1, 0,
						length.x + 1,
						getSize().height - 1);
					g.setColor(currentForeground);
					drawRadar(g, length.x + 3, start.y,
						length.x - start.x,
						length.y - start.y);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void moveNoPieces() {
		((MballFrame) frame).callback(MballInterface.ACTION_ILLEGAL);
	}

	void swapPieces(int wedge1, int wedge2) {
		MballLoc tempLoc = new MballLoc(0, 0);
		int band;

		if (wedge1 == wedge2) {
			for (band = 0; band < bands / 2; band++) {
				tempLoc.wedge = wedgeLoc[wedge1 + wedges * band].wedge;
				tempLoc.direction = wedgeLoc[wedge1 + wedges * band].direction;
				wedgeLoc[wedge1 + wedges * band].wedge = wedgeLoc[wedge1 +
					wedges * (bands - 1 - band)].wedge;
				wedgeLoc[wedge1 + wedges * band].direction = wedgeLoc[wedge1 +
					wedges * (bands - 1 - band)].direction;
				wedgeLoc[wedge1 + wedges * (bands - 1 - band)].wedge =
					tempLoc.wedge;
				wedgeLoc[wedge1 + wedges * (bands - 1 - band)].direction =
					tempLoc.direction;
			}
			for (band = 0; band < bands; band++)
				wedgeLoc[wedge1 + wedges * band].direction = (wedgeLoc[wedge1 +
					wedges * band].direction == UP) ? DOWN : UP;
			drawWedge(wedge1);
		} else {
			for (band = 0; band < bands; band++) {
				tempLoc.wedge = wedgeLoc[wedge1 + wedges * band].wedge;
				tempLoc.direction = wedgeLoc[wedge1 + wedges * band].direction;
				wedgeLoc[wedge1 + wedges * band].wedge = wedgeLoc[wedge2 +
					wedges * (bands - 1 - band)].wedge;
				wedgeLoc[wedge1 + wedges * band].direction = (wedgeLoc[wedge2 +
					wedges * (bands - 1 - band)].direction == UP) ?
					DOWN : UP;
				wedgeLoc[wedge2 + wedges * (bands - 1 - band)].wedge =
					tempLoc.wedge;
				wedgeLoc[wedge2 + wedges * (bands - 1 - band)].direction =
					(tempLoc.direction) == UP ? DOWN : UP;
			}
			drawWedge(wedge1);
			drawWedge(wedge2);
		}
	}

	void movePieces(int wedge, int band, int direction) {
		int i;

		if (direction == CW || direction == CCW) { /* rotate band */
			int newI;
			MballLoc temp1Loc = new MballLoc(0, 0);
			MballLoc temp2Loc = new MballLoc(0, 0);

			for (i = 0; i < wedges; i++) {
				newI = (direction == CW) ? i : wedges - 1 - i;
				if (newI == ((direction == CW) ? 0 : wedges - 1)) {
					temp1Loc.wedge = wedgeLoc[newI + wedges * band].wedge;
					temp1Loc.direction = wedgeLoc[newI +
						wedges * band].direction;
					wedgeLoc[newI + wedges * band].wedge =
						wedgeLoc[((direction == CW) ?
						wedges - 1 : 0) + wedges * band].wedge;
					wedgeLoc[newI + wedges * band].direction =
						wedgeLoc[((direction == CW) ?
						wedges - 1 : 0) + wedges * band].direction;
				} else {
					temp2Loc.wedge = temp1Loc.wedge;
					temp2Loc.direction = temp1Loc.direction;
					temp1Loc.wedge = wedgeLoc[newI + wedges * band].wedge;
					temp1Loc.direction = wedgeLoc[newI +
						wedges * band].direction;
					wedgeLoc[newI + wedges * band].wedge =
						temp2Loc.wedge;
					wedgeLoc[newI + wedges * band].direction =
						temp2Loc.direction;
				}
				drawSector(newI + wedges * band, 0);
			}
		} else { /* flip */
			int sphereDir = mapDirToWedge[(wedges - MballInterface.MIN_WEDGES) / 2][direction];
			int flipOffset = wedges / 2;
			int wedge1, wedge2;

			for (i = 0; i < wedges / 2; i++)
				if (wedge == i + sphereDir)
					flipOffset = 0;
			for (i = 0; i < (wedges + 2) / 4; i++) {
				wedge1 = (i + sphereDir + flipOffset) % wedges;
				wedge2 = (wedges / 2 - 1 - i + sphereDir + flipOffset) %
					wedges;
				swapPieces(wedge1, wedge2);
			}
		}
	}

	void moveControlCb(int wedge, int direction) {
		if (direction > COORD) {
			for (int band = 0; band < bands; band++) {
				movePieces(wedge, band, direction);
				((MballFrame) frame).callback(
					MballInterface.ACTION_CONTROL);
			}
		} else {
			movePieces(0, 0, direction);
			((MballFrame) frame).callback(
				MballInterface.ACTION_CONTROL);
			movePieces(wedges / 2, 0, direction);
			((MballFrame) frame).callback(
				MballInterface.ACTION_CONTROL);
		}
	}

	void movePuzzle(int sector, int direction, boolean control) {
		int wedge = sector % wedges;
		int band = sector / wedges;

		if (control) {
			moveControlCb(wedge, direction);
		} else {
			movePieces(wedge, band, direction);
			((MballFrame) frame).callback(
				MballInterface.ACTION_MOVED);
		}
		undo.setMove(direction, (control) ? 1 : 0,
			wedge + wedges * band);
		redo.flushMoves();
		if (((MballFrame) frame).getToggleSound())
			((MballFrame) frame).playMoveAudio();

	}

	void movePuzzleDelay(int wedge, int band, int direction,
			boolean control, int fast) {
		movePuzzle(wedge + wedges * band, direction, control);
		try {
			Thread.sleep(64 * delay / fast);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	int positionToSector(int positionX, int positionY) {
		double angle, radius;
		int wedge, band;
		int x = positionX - puzzleOffset.x;
		int y = positionY - puzzleOffset.y;

		if (vertical && y > viewLength - 1) {
			y -= (viewLength - 1);
			currentView = DOWN;
		} else if (!vertical && x > viewLength - 1) {
			x -= (viewLength - 1);
			currentView = DOWN;
		} else
			currentView = UP;
		x -= (wedgeLength + 1) / 2;
		y -= (wedgeLength + 1) / 2;
		radius = Math.sqrt((double) x * x + y * y);
		if (y >= 0)
			angle = Math.atan2(-x, y) + Math.PI;
		else if (x < 0)
			angle = 2 * Math.PI - Math.atan2(-x, -y);
		else
			angle = -Math.atan2(-x, -y);
		if (perspective) {
			band = (bands - 1) >> 1;
			for (int i = 0; i < ((bands - 1) >> 1); i++) {
				/*if (i == bands >> 1 && bands % 2 == 1 &&
						radius < ((wedgeLength + 1) >> 1)) {
					band = i;
					break;
				} else*/ if (radius < wedgeLength * Math.sin(Math.PI *
						(i + 1) / bands) / 2.0) {
					band = i;
					break;
				}
			}
		} else {
			band = (int) (radius * bands / wedgeLength);
			if (band > (bands - 1) >> 1)
				band--;
		}
		wedge = (int) (angle * wedges / (2.0 * Math.PI));
		if (currentView == DOWN) {
			if (vertical)
				wedge = ((3 * wedges) / 2 - 1 - wedge) % wedges;
			else
				wedge = (wedges - 1 - wedge) % wedges;
			band = bands - 1 - band;
		}
		if (radius > wedgeLength / 2 + delta)
			return -1;
		return (wedge + wedges * band);
	}

	boolean positionSector(int x, int y, int direction) {
		currentSector = positionToSector(x, y);
		currentDirection = direction;
		if ((currentDirection == CW || currentDirection == CCW) &&
				currentSector < 0)
			return false;
		if (currentView == DOWN) {
			if (currentDirection == CCW)
				currentDirection = CW;
			else if (currentDirection == CW)
				currentDirection = CCW;
			else if (currentDirection < COORD)
				currentDirection =
					(COORD - currentDirection) % COORD;
		}
		return (currentDirection == CW || currentDirection == CCW ||
				mapDirToWedge[(wedges -
				MballInterface.MIN_WEDGES) /
				2][currentDirection] != CUTS);
	}

	void movePuzzleInput(int x, int y, int direction,
			boolean control) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (checkSolved() && !practice && !control) {
			moveNoPieces();
			return;
		}
		if (!positionSector(x, y, direction))
			return;
		movePuzzle(currentSector, currentDirection, control);
		if (!control && checkSolved()) {
			((MballFrame) frame).callback(
				MballInterface.ACTION_SOLVED);
		}
	}

	void selectSectors() {
		if (!nextControl && !practice && checkSolved())
			moveNoPieces();
		else if (nextSector >= 0) {
			int opp = (currentSector + wedges / 2) % wedges;
			int currentWedge = currentSector % wedges;
			int currentBand = currentSector / wedges;
			int nextWedge = nextSector % wedges;
			int nextBand = nextSector / wedges;

			if (nextBand == currentBand) {
				if (nextWedge == currentWedge)
					return;
				if (nextShift || (opp != (nextWedge + 1) % wedges &&
				    nextWedge != (opp + 1) % wedges)) {
					/*((MballFrame) frame).callback(
					  MballInterface.ACTION_AMBIGUOUS);*/
				  int diff = (currentWedge - nextWedge +
				    wedges) % wedges;
				  if (diff > wedges / 2)
				    for (int i = 0; i < wedges - diff; i++)
				      movePuzzle(nextWedge + wedges * nextBand,
					CW, nextControl);
				  else
				    for (int i = 0; i < diff; i++)
				      movePuzzle(nextWedge + wedges * nextBand,
					CCW, nextControl);
				  started = true;
				  if (!nextControl && checkSolved()) {
				    ((MballFrame) frame).callback(
				      MballInterface.ACTION_SOLVED);
				  }
				  return;
				}
			}
			if (nextWedge == currentWedge && wedges > 2) {
				((MballFrame) frame).callback(
					MballInterface.ACTION_AMBIGUOUS);
			} else if (opp == (nextWedge + 1) % wedges) {
				started = true;
				movePuzzle(nextWedge + wedges * nextBand,
					mapWedgeToDir[(wedges -
					MballInterface.MIN_WEDGES) / 2]
					[currentWedge], nextControl);
				if (!nextControl && checkSolved()) {
				    ((MballFrame) frame).callback(
				      MballInterface.ACTION_SOLVED);
				}
			} else if (nextWedge == (opp + 1) % wedges) {
				started = true;
				movePuzzle(nextWedge + wedges * nextBand,
					mapWedgeToDir[(wedges -
					MballInterface.MIN_WEDGES) / 2]
					[nextWedge], nextControl);
				if (!nextControl && checkSolved()) {
				    ((MballFrame) frame).callback(
				      MballInterface.ACTION_SOLVED);
				}
			} else {
				moveNoPieces();
			}
		}
	}

	void resetPieces() {
		solve.setFlag(false);
		wedgeBands = bands * wedges;
		wedgeLoc = new MballLoc[wedgeBands];
		startLoc = new MballLoc[wedgeBands];
		for (int sector = 0; sector < wedgeBands; sector++) {
			startLoc[sector] = new MballLoc(sector % wedges, DOWN);
			wedgeLoc[sector] = new MballLoc(sector % wedges, DOWN);
		}
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentSector = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void practicePieces() {
		((MballFrame) frame).callback(MballInterface.ACTION_PRACTICE);
	}

	void randomizePieces() {
		if (practice)
			practicePieces();
		if (bands > 1 || wedges > 1) {
			((MballFrame) frame).callback(
				MballInterface.ACTION_RESET);
			lastDirection = RIGHT;
			randomPaints = java.lang.Math.min(wedges * (bands + 1),
				MAX_ITERATIONS) + generator.nextInt(2);
			repaint();
		} else if (checkSolved())
			((MballFrame) frame).callback(
				MballInterface.ACTION_SOLVED);
	}

	void randomizingPieces() {
		int wedge = generator.nextInt(wedges);
		int band = generator.nextInt(bands);

		do
			currentDirection = generator.nextInt(2 * COORD);
		while (currentDirection < COORD && mapDirToWedge[(wedges -
			MballInterface.MIN_WEDGES) / 2][currentDirection] ==
			CUTS);
		if (currentDirection >= COORD) {
			if (currentDirection - COORD < CUTS)
				currentDirection = CW;
			else
				currentDirection = CCW;
		}
		if ((currentDirection + COORD / 2) % COORD +
				currentDirection / COORD != lastDirection) {
			movePuzzle(wedge + wedges * band, currentDirection,
				false);
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((MballFrame) frame).callback(
				MballInterface.ACTION_RANDOMIZE);
			started = false;
			currentSector = -1;
			if (checkSolved())
				((MballFrame) frame).callback(
					MballInterface.ACTION_SOLVED);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
	}

	public boolean getCheat() {
		return cheat;
	}

	boolean setPuzzleOrient(boolean newValue) {
		if (orient != newValue) {
			orient = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePractice(boolean newValue) {
		if (practice != newValue) {
			practice = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleBands(int newValue) {
		if (bands != newValue && newValue >= MballInterface.MIN_BANDS) {
			bands = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleWedges(int newValue) {
		if (wedges != newValue && newValue >= MballInterface.MIN_WEDGES
				&& newValue <= MballInterface.MAX_WEDGES &&
				((newValue & 1) == 0)) {
			wedges = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((MballFrame) frame).callback(MballInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((MballFrame) frame).callback(MballInterface.ACTION_SPEED);
	}

	boolean setPuzzlePerspective(boolean newValue) {
		if (perspective != newValue) {
			perspective = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}
	void showMessage(String msg) {
		((MballFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		int tempLength;

		coreSize = new Point(getSize().width, getSize().height);
		delta = 4;
		vertical = (getSize().height >= getSize().width);
		if (vertical)
			tempLength = java.lang.Math.min(getSize().height / 2,
				getSize().width);
		else
			tempLength = java.lang.Math.min(getSize().height,
				getSize().width / 2);
		mballLength = java.lang.Math.max((tempLength - delta + 1) /
			wedges, 0);
		wedgeLength = wedges * mballLength;
		viewLength = wedgeLength + delta;
		if (vertical) {
			puzzleSize = new Point(viewLength - 1,
				2 * viewLength - delta - 2);
		} else {
			puzzleSize = new Point(2 * viewLength - delta - 2,
				viewLength - 1);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		mballLength = wedgeLength / (2 * wedges) - delta - 1;
		dr = wedges;
	}

	public void initializePuzzle() {
		//for (int wedge = 0; wedge < MballInterface.MAX_WEDGES; wedge++)
		//	mballLoc[wedge] = NULL;
		checkPieces();
		undo = new MballStack();
		redo = new MballStack();
		solve = new MballSolve(frame, this);
		resetPieces();
		generator = new Random(System.nanoTime());
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		drawAllPieces();
		solve.init();
	}

	void selectPuzzle(boolean control, int x, int y) {
		currentSector = positionToSector(x, y);
		if (currentSector >= 0) {
			if (control || practice || !checkSolved()) {
				selectPaint = true;
				repaint();
				return;
			}
			currentSector = -1;
		}
		currentSector = -1;
	}

	void releasePuzzle(boolean shift, boolean control, int x, int y) {
		if (currentSector == -1)
			return;
		nextShift = shift;
		nextControl = control;
		nextSector = positionToSector(x, y);
		releasePaint = true;
		repaint();
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizePieces();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle() {
		this.stringSave = ((MballFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, bandsValue, wedgesValue, moves;
			boolean orientValue, practiceValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			wedgesValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (wedgesValue >= MballInterface.MIN_WEDGES &&
					wedgesValue <= MballInterface.MAX_WEDGES &&
					(wedgesValue & 1) == 0) {
				if (wedgesValue != this.wedges) {
					int reason = (wedgesValue -
						MballInterface.MIN_WEDGES) / 2 +
						MballInterface.ACTION_WEDGE2;

					((MballFrame) frame).callback(reason);
				}
			} else {
				showMessage(source +
					" corrupted: wedges " + wedgesValue +
					" should be even and between " +
					MballInterface.MIN_WEDGES +
					" and " + MballInterface.MAX_WEDGES +
					" inclusive");
				return;
			}
			lineData = token[++line];
			bandsValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (bandsValue >= MballInterface.MIN_BANDS) {
				for (i = this.bands; i < bandsValue; i++) {
					((MballFrame) frame).callback(
						MballInterface.ACTION_INCREMENT);
				}
				for (i = this.bands; i > bandsValue; i--) {
					((MballFrame) frame).callback(
						MballInterface.ACTION_DECREMENT);
				}
			} else {
				showMessage(source +
					" corrupted: bands " + bandsValue +
					" should be at least " +
					MballInterface.MIN_BANDS);
				return;
			}
			lineData = token[++line];
			orientValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orient != orientValue)
				((MballFrame) frame).callback(
					MballInterface.ACTION_ORIENTIZE);
			lineData = token[++line];
			practiceValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practice != practiceValue)
				((MballFrame) frame).callback(
					MballInterface.ACTION_PRACTICE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((MballFrame) frame).callback(
				MballInterface.ACTION_RESTORE);
			setStartPosition();
			allWedgesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": wedges " + wedgesValue +
				", bands " + bandsValue + ", orient " + orientValue +
				", practice " + practiceValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("wedges").append(symbol).append(space)
			.append(wedges).append(newline);
		sb.append("bands").append(symbol).append(space)
			.append(bands).append(newline);
		sb.append("orient").append(symbol).append(space)
			.append((orient) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
			.append((practice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((MballFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentControl = (undo.readMoveControl() != 0);
		currentSector = undo.readMoveSector();
		redo.setMove(currentDirection, (currentControl) ? 1 : 0,
			currentSector);
		// move in opposite direction
		currentDirection = (currentDirection < COORD) ?
			currentDirection : 3 * COORD - currentDirection;
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((MballFrame) frame).callback(
				MballInterface.ACTION_UNDO);
		} /* else currentDirection = currentSector = -1; */
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentControl = (redo.readMoveControl() != 0);
		currentSector = redo.readMoveSector();
		undo.setMove(currentDirection, (currentControl) ? 1 : 0,
			currentSector);
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((MballFrame) frame).callback(
				MballInterface.ACTION_REDO);
		} /* else currentDirection = currentSector = -1; */
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetPieces();
		drawAllPieces();
		allWedgesPaint = true;
		repaint();
		((MballFrame) frame).callback(MballInterface.ACTION_RESET);
	}

	void solvePuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		if (wedges <= 8 || bands == 1) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solvePieces();
		} else {
			showMessage("Auto-solver: sorry, only implemented for\nwedges <= 8 or 1 band.");
		}
	}

	void incrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_INCREMENT);
	}

	void decrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		if (bands <= MballInterface.MIN_BANDS)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_DECREMENT);
	}

	void wedge2ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE2);
	}

	void wedge4ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE4);
	}

	void wedge6ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE6);
	}

	void wedge8ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE8);
	}

	void wedge10ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE10);
	}

	void wedge12ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_WEDGE12);
	}

	void orientizePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_ORIENTIZE);
	}

	void practicePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		practicePieces();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void perspectivePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MballFrame) frame).callback(
			MballInterface.ACTION_PERSPECTIVE);
	}
	void printPositions() {
		for (int i = 0; i < wedgeBands; i++) {
			int pos = startLoc[i].wedge;
			int dir = startLoc[i].direction;

			if (pos < 100)
				System.out.print(" ");
			if (pos < 10 && pos >= 0)
				System.out.print(" ");
			System.out.print(pos + "," + dir);
			if (i % wedges == wedges - 1) {
				System.out.println();
			} else {
				System.out.print(" ");
			}
		}
		System.out.println();
	}

	void getStartPosition() {
		for (int i = 0; i < wedgeBands; i++) {
			startLoc[i].wedge = wedgeLoc[i].wedge;
			startLoc[i].direction = wedgeLoc[i].direction;
		}
	}

	void setStartPosition() {
		for (int i = 0; i < wedgeBands; i++) {
			wedgeLoc[i].wedge = startLoc[i].wedge;
			wedgeLoc[i].direction = startLoc[i].direction;
		}
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";

			++line; //read startingPosition:
			for (int band = 0; band < bands; band++) {
				line++;
				if (line >= token.length) {
					System.out.println("Short data in StartPosition");
					return -1;
				}
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int wedge = 0; wedge < wedges; wedge++) {
					int num = Integer.parseInt(st[2 * wedge + offset].trim());

					startLoc[wedge + band * wedges].wedge = num;
					num = Integer.parseInt(st[2 * wedge + offset + 1].trim());
					startLoc[wedge + band * wedges].direction = num;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading (StartPosition)");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentControl = (Integer.parseInt(st[1].trim()) != 0);
				currentSector = Integer.parseInt(st[2].trim());
				{
					scanPaint = true;
					undo.setMove(currentDirection,
						(currentControl) ? 1 : 0,
						currentSector);
					redo.flushMoves();
					paint(g);
					if (!currentControl)
						((MballFrame) frame).callback(
						MballInterface.ACTION_MOVED);
				} /* else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					return false;
				} */
			}
			if (l < moves) {
				showMessage("Corrupt input (scanMoves), only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading (scanMoves)");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < wedgeBands; i++) {
			int temp = startLoc[i].wedge;
			if (temp < 100)
				sb.append(space);
			if (temp < 10 && temp >= 0)
				sb.append(space);
			sb.append(space).append(temp);
			temp = startLoc[i].direction;
			if (temp < 10 && temp >= 0)
				sb.append(space);
			sb.append(space).append(temp);
			if (((i + 1) % wedges) == 0) {
				sb.append(newline);
			}
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction, sector, control;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("dir").append(space)
			.append("control").append(space)
			.append("sector").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			control = undo.readMoveControl();
			sector = undo.readMoveSector();
			redo.setMove(direction, control, sector);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			control = redo.readMoveControl();
			sector = redo.readMoveSector();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(control).append(space)
				.append(sector).append(newline);
			undo.setMove(direction, control, sector);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allWedgesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allWedgesPaint) {
			drawAllPieces();
			allWedgesPaint = false;
		}
		if (selectPaint) {
			if (currentSector != -1) {
				drawSector(currentSector, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentSector != -1) {
				drawSector(currentSector, 0);
				selectSectors();
				currentSector = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (currentControl)
				moveControlCb(currentSector, currentDirection);
			else {
				movePieces(currentSector % wedges,
					currentSector / wedges,
					currentDirection);
				if (stackPaint && checkSolved()) {
					((MballFrame) frame).callback(
						MballInterface.ACTION_SOLVED);
				}

			}
			currentDirection = currentSector = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}

/**
 * The <code>MballLoc</code> class maintains the locations of
 * the pieces.
 */

class MballLoc {
	int	 wedge;
	int	 direction;

	public MballLoc(int wedge, int direction) {
		this.wedge = wedge;
		this.direction = direction;
	}
}
