package com.sillycycle.bagleyd.mball;

/*
 * @(#)MballFrame.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.Colour;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>MballFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class MballFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_BANDS = 6;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "mball.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Practice", "Randomize", "Auto-solve",
		"Increment", "Decrement", "Orientize",
		"Speed >", "Slow <", "Sound @", "Perspective",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'p', 'z', 's',
		'i', 'd', 'o', '>', '<', '@', 't',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 460;
	int heightValue = 230;
	static int xValue = 0;
	static int yValue = 0;
	String monoString = "YGSBCMROPTLI";
	Color[] wedgeColor = new Color[] {
		Color.yellow, Color.green, Colour.seaGreen,
		Color.blue, Color.cyan, Color.magenta, Color.red,
		Color.orange, Color.pink, Colour.lightTan,
		Colour.lightSteelBlue, Colour.indianRed
	};
	Color borderColor = Color.black;
	Color fgColor = Color.cyan;
	Color bgColor = Colour.steelBlue;
	boolean sound = false;
	boolean controlValue = true;
	boolean orientValue = MballInterface.DEFAULT_ORIENT;
	boolean practiceValue = MballInterface.DEFAULT_PRACTICE;
	int bandsValue = MballInterface.DEFAULT_BANDS;
	int wedgesValue = MballInterface.DEFAULT_WEDGES;
	int baseValue = MballInterface.DEFAULT_BASE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean perspectiveValue = MballInterface.DEFAULT_PERSPECTIVE;
	boolean debug = false;
	JButton button = null;
	protected JPopupMenu popup = null;
	JPanel topPanel = new JPanel();
	JComboBox<Object> wedgeComboBox = new JComboBox<>();
	JCheckBox orientCheckBox = new JCheckBox("Oriented");
	JCheckBox practiceCheckBox = new JCheckBox("Practice");
	Mball puzzle;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Masterball! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][][] puzzleRecord = null;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner bandChanger = null, speedChanger = null;
	//Slider bandChanger = null, speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle.stringSave = s;
	}

	public void processKey(char key) {
		puzzle.processKey(key);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {

			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		for (int i = 0; i < wedgeColor.length; i++)
			if (i >= 10)
				wedgeColor[i] = getOpt("wedgeColor" +
					('A' + i - 10), wedgeColor[i]);
			else
				wedgeColor[i] = getOpt("wedgeColor" + i,
					wedgeColor[i]);
		borderColor = getOpt("borderColor", borderColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		wedgesValue = getOpt("wedges", wedgesValue);
		if ((wedgesValue & 1) != 0 ||
				wedgesValue < MballInterface.MIN_WEDGES ||
				wedgesValue > MballInterface.MAX_WEDGES)
			wedgesValue = MballInterface.DEFAULT_WEDGES;
		bandsValue = getOpt("bands", bandsValue);
		if (bandsValue < MballInterface.MIN_BANDS)
			bandsValue = MballInterface.DEFAULT_BANDS;
		orientValue = getOpt("orient", orientValue);
		practiceValue = getOpt("practice", practiceValue);
		baseValue = getOpt("base", baseValue);
		if (baseValue < MballInterface.MIN_BASE ||
				baseValue > MballInterface.MAX_BASE)
			baseValue = MballInterface.DEFAULT_BASE;
		delayValue = getOpt("delay", delayValue);
		perspectiveValue = getOpt("perspective", perspectiveValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Mball", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Mball - the original puzzle, Masterball, has 8 wedges on a" + NL +
"sphere (longitudinal cuts), with each wedge divided into" + NL +
"divided into 4 segments (latitudinal cuts).  There are" + NL +
"essentially 3 varieties: Geomaster 8 colors (beachball, default" + NL +
"colors in this puzzle), Duomaster 2 colors (black and white" + NL +
"beachball), and a variety of picture Masterballs.  For some" + NL +
"reason, they say the Geomaster is easier than the Duomaster." + NL +
"The picture Masterballs are the hardest since all the pieces" + NL +
"have a set solved position.  On the Duo and Geo Masterballs" + NL +
"pieces could be swapped or in a different order and still be" + NL +
"in a solved position.  Zurick University's Mathematics faculty" + NL +
"has calculated that Masterball's 32 segments can be arranged" + NL +
"in 355,682,548,566,633,480,192,000,000 different possible" + NL +
"combinations.  Masterball was invented by Dr. Geza Gyovai, an" + NL +
"Hungarian Engineer, manufactured by Whole Systems Design, Inc.." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Make sure you randomize the puzzle first before trying to move the" + NL +
"wedges for the first time." + NL +
" " + NL +
"Press \"mouse-left\" button to move a wedge.  Release \"mouse-left\"" + NL +
"button on another piece and the pieces will turn towards where" + NL +
"the mouse button is released.  Usually, a click and release" + NL +
"on the same wedge is ambiguous and the puzzle will not turn." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"Z\" or \"z\" keys, to" + NL +
"randomize the puzzle (this must be done first to set a new record)." + NL +
"One must double click on \"mouse-right\" if the puzzle is currently" + NL +
"being worked on." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear the puzzle." + NL +
" " + NL +
"Press \"P\" or \"p\" keys, to toggle the practice mode (in practice" + NL +
"mode the record should say \"practice\").  This is good for" + NL +
"learning moves and experimenting." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to start auto-solver.  Only works with 2, 4," + NL +
"or 8 wedges or 1 band." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the orient mode.  (Default is off)." + NL +
"If on, puzzle has to solved in clockwise order, besides getting" + NL +
"all the wedges to be the same color." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to increase the number of bands." + NL +
" " + NL +
"Press \"D\" or \"d\" keys to decrease the number of bands." + NL +
" " + NL +
"Press \"2\", \"4\", \"6\", \"8\", \"0\", or \"=\" keys (not the" + NL +
"keypad 2, 4, 6, 8, 0) to change to 2, 4, 6, 8, 10, or 12 wedges," + NL +
"wedges, respectively.  Note: if there were odd number of wedges," + NL +
"there would be no 180 degree turn and therefore the puzzle would" + NL +
"inoperable." + NL +
" " + NL +
"Press \"T\" or \"t\" keys to toggle the perspective look." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the wedges." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the wedges." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the mouse." + NL +
"Key pad (better with Num Lock on) is defined for Mball as:" + NL +
"N / *  Upper Upper Left, Counterclockwise, Upper Upper Right" + NL +
" " + NL +
"7 8 9  Upper Left, Up, Upper Right" + NL +
"   ^" + NL +
"4<5>6  Left, Clockwise, Right" + NL +
"   v" + NL +
"1 2 3  Lower Left, Down, Lower Right" + NL +
" " + NL +
"0    . Lower Lower Left, Lower Lower Right" + NL +
"Use the control key and the left mouse button, keypad, or arrow" + NL +
"keys to move the whole puzzle.  This is not recorded as a turn." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"http://wsd.com/masterball" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java MballFrame Version 8.7.2" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = MballFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = MballFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playMoveAudio() {
		playClip(MballInterface.SOUNDS[MballInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle == null)
			return;
		puzzle.setBounds(0, 0, width, height);
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int i = 0; i < puzzleRecord.length; i++) {
			for (int j = 0; j < puzzleRecord[i].length; j++) {
				for (int k = 0; k < puzzleRecord[i][j].length; k++) {
					scoreString.append(puzzleRecord[i][j][k]);
					if (k < puzzleRecord[i][j].length - 1)
						scoreString.append(" ");
				}
				if (j < puzzleRecord[i].length - 1
						|| i < puzzleRecord.length - 1)
					scoreString.append(NL);
			}
			if (i < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[2]
			[(MballInterface.MAX_WEDGES - MballInterface.MIN_WEDGES) / 2 + 1]
			[MAX_BANDS - MballInterface.MIN_BANDS + 1];
		for (int i = 0; i < puzzleRecord.length; i++) {
		  for (int j = 0; j < puzzleRecord[i].length; j++) {
		    for (int k = 0; k < puzzleRecord[i][j].length; k++) {
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[i][j][k] =
					MballScores.PUZZLE_SCORES[i][j][k];
			} else {
				puzzleRecord[i][j][k] =
					Integer.parseInt(st[field++]);
			}
		    }
		  }
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(int wedges, int bands,
			boolean orient, boolean practice) {
		int i = (orient) ? 1 : 0;
		int j = (wedges - MballInterface.MIN_WEDGES) / 2;
		int k = bands - MballInterface.MIN_BANDS;

		if (puzzleRecord == null)
			return;
		if (practice) {
			recordValueLabel.setText("PRACTICE");
		} else if (bands > MAX_BANDS) {
			recordValueLabel.setText("NONE");
		} else if (puzzleRecord[i][j][k] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[i][j][k]));
		}
	}

	boolean handleSolved(int counter, int wedges, int bands,
			boolean orient, boolean practice) {
		int i = (orient) ? 1 : 0;
		int j = (wedges - MballInterface.MIN_WEDGES) / 2;
		int k = bands - MballInterface.MIN_BANDS;

		if (puzzleRecord != null &&
				wedges <= MballInterface.MAX_WEDGES &&
				(counter < puzzleRecord[i][j][k] ||
				puzzleRecord[i][j][k] <= NEVER)) {
			readScoreRecords(); /* Maybe its been updated by another */
			puzzleRecord[i][j][k] = counter;
			if (wedges < 4 || bands < 2) {
				puzzleRecord[i][j][k] = counter;
			}
			writeScore();
			printRecord(wedges, bands,
				orient, practice);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
		}
	}

	class BandChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = bandsValue;

			bandsValue = bandChanger.getValue();
			if (oldValue != bandsValue) {
				puzzle.setPuzzleBands(bandsValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(wedgesValue, bandsValue,
					orientValue, practiceValue);
			}
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	class ComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int wedgesIndex = wedgeComboBox.getSelectedIndex() * 2
				+ MballInterface.MIN_WEDGES;

			if (wedgesIndex != -1 && wedgesValue != wedgesIndex) {
				wedgesValue = wedgesIndex;
				puzzle.setPuzzleWedges(wedgesValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(wedgesValue, bandsValue,
					orientValue, practiceValue);
			}
		}
	}

	class OrientItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = orientValue;

			orientValue = orientCheckBox.isSelected();
			if (oldValue != orientValue) {
				puzzle.setPuzzleOrient(orientValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(wedgesValue, bandsValue,
					orientValue, practiceValue);
			}
		}
	}

	class PracticeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = practiceValue;

			practiceValue = practiceCheckBox.isSelected();
			if (oldValue != practiceValue) {
				puzzle.setPuzzlePractice(practiceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(wedgesValue, bandsValue,
					orientValue, practiceValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public void callback(int reason) {
		messageDisplay = "                                                                                ";
		switch(reason) {
		case MballInterface.ACTION_RESTORE:
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			movesDisplay = 0;
			break;
		case MballInterface.ACTION_RESET:
			movesDisplay = 0;
			break;
		case MballInterface.ACTION_AMBIGUOUS:
			messageDisplay = "Ambiguous move";
			break;
		case MballInterface.ACTION_ILLEGAL:
			if (practiceValue || puzzle.checkStarted())
				messageDisplay = "Illegal move";
			else
				messageDisplay = "Randomize to start";
			break;
		case MballInterface.ACTION_MOVED:
			movesDisplay++;
			break;
		case MballInterface.ACTION_CONTROL:
			break;
		case MballInterface.ACTION_SOLVED:
			if (practiceValue)
				movesDisplay = 0;
			else {
				if (puzzle.getCheat())
					messageDisplay = "No cheating!!"; //username
				else if (handleSolved(movesDisplay, wedgesValue,
						bandsValue, orientValue,
						practiceValue))
					messageDisplay = "Best solution, congratulations!!"; //username
				else
					messageDisplay = "Solved!";
			}
			//showMessage(messageDisplay);
			break;
		case MballInterface.ACTION_COMPUTED:
			break;
		case MballInterface.ACTION_PRACTICE:
			movesDisplay = 0;
			practiceValue = !practiceValue;
			if (!practiceValue)
				messageDisplay = "Randomize to start";
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			puzzle.setPuzzlePractice(practiceValue);
			practiceCheckBox.setSelected(practiceValue);
			break;
		case MballInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			practiceValue = false;
			puzzle.setPuzzlePractice(practiceValue);
			break;
		case MballInterface.ACTION_INCREMENT:
			movesDisplay = 0;
			bandsValue++;
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			puzzle.setPuzzleBands(bandsValue);
			if (bandChanger != null)
				bandChanger.setValue(bandsValue);
			break;
		case MballInterface.ACTION_DECREMENT:
			movesDisplay = 0;
			bandsValue--;
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			puzzle.setPuzzleBands(bandsValue);
			if (bandChanger != null)
				bandChanger.setValue(bandsValue);
			break;
		case MballInterface.ACTION_ORIENTIZE:
			movesDisplay = 0;
			orientValue = !orientValue;
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			puzzle.setPuzzleOrient(orientValue);
			orientCheckBox.setSelected(orientValue);
			break;
		case MballInterface.ACTION_WEDGE2:
		case MballInterface.ACTION_WEDGE4:
		case MballInterface.ACTION_WEDGE6:
		case MballInterface.ACTION_WEDGE8:
		case MballInterface.ACTION_WEDGE10:
		case MballInterface.ACTION_WEDGE12:
			movesDisplay = 0;
			wedgesValue = 2 * (reason -
				MballInterface.ACTION_WEDGE2) +
				MballInterface.MIN_WEDGES;
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			puzzle.setPuzzleWedges(wedgesValue);
			wedgeComboBox.setSelectedItem(
				MballInterface.wedgeStrings[(wedgesValue -
					MballInterface.MIN_WEDGES) / 2]);
			break;
		case MballInterface.ACTION_UNDO:
			movesDisplay--;
			break;
		case MballInterface.ACTION_REDO:
			movesDisplay++;
			break;
		case MballInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				delay = puzzle.getDelay();
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					puzzle.setDelay(delay);
				}
			}
			return;
		case MballInterface.ACTION_PERSPECTIVE:
			movesDisplay = 0;
			perspectiveValue = !perspectiveValue;
			puzzle.setPuzzlePerspective(perspectiveValue);
			break;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(
			movesDisplay));
	}

	public MballFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel checkBoxPanel = new JPanel();
		JPanel comboBoxPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel canvasPanel = new JPanel();
		int wedge;

		try {
			mediumIcon = getIcon(MballInterface.ICONS[MballInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(MballInterface.ICONS[MballInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			MballInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		setBackground(Color.white);
		setLayout(new BorderLayout());
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			// Allows menus to be on top of canvas
			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Practice");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Increment");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_I);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Decrement");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_D);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Orientize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sound @");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Perspective");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_T);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move wedge"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Randomize"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(8);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			for (wedge = 0; wedge <= (MballInterface.MAX_WEDGES -
					MballInterface.MIN_WEDGES) / 2; wedge++) {
				wedgeComboBox.addItem(
					MballInterface.wedgeStrings[wedge]);
			}
			wedgeComboBox.setSelectedItem(MballInterface.
				wedgeStrings[(wedgesValue -
				MballInterface.MIN_WEDGES) / 2]);
			comboBoxPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			comboBoxPanel.add(wedgeComboBox);
			comboBoxPanel.add(new JLabel("Wedges"));
			bandChanger = new Spinner("Bands",
				bandsValue, MballInterface.MIN_BANDS,
				((bandsValue > MAX_BANDS) ? bandsValue : MAX_BANDS));
			/*bandChanger = new Slider("Bands",
				bandsValue, MballInterface.MIN_BANDS,
				((bandsValue > MAX_BANDS) ? bandsValue : MAX_BANDS));*/
			sliderPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			sliderPanel.add(comboBoxPanel);
			sliderPanel.add(bandChanger);
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			speedChanger.setWidth(150);*/
			checkBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			checkBoxPanel.add(orientCheckBox);
			orientCheckBox.setSelected(orientValue);
			checkBoxPanel.add(practiceCheckBox);
			practiceCheckBox.setSelected(practiceValue);
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			speedChanger.addChangeListener(new SpeedChangeListener());
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			topPanel.add(sliderPanel);
			topPanel.add(speedChanger);
			topPanel.add(checkBoxPanel);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		puzzle = new Mball(this,
			orientValue, practiceValue,
			wedgesValue, bandsValue,
			baseValue, false, false, monoString,
			wedgeColor[0], wedgeColor[1], wedgeColor[2],
			wedgeColor[3], wedgeColor[4], wedgeColor[5],
			wedgeColor[6], wedgeColor[7], wedgeColor[8],
			wedgeColor[9], wedgeColor[10], wedgeColor[11],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue,
			perspectiveValue);
		canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvasPanel.add(puzzle);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(wedgesValue, bandsValue,
				orientValue, practiceValue);
			bandChanger.addChangeListener(new BandChangeListener());
			wedgeComboBox.addActionListener(new ComboBoxActionListener());
			orientCheckBox.addItemListener(new OrientItemListener());
			practiceCheckBox.addItemListener(new PracticeItemListener());
		}
		canvasPanel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : MballInterface.ICONS) {
			URL url = MballFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-wedgeColor{0|1|2|3|4|5|6|7|8|9|A|B}={int}]" + NL +
"\t[-borderColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}] [-[no]perspective={bool}]" + NL +
"\t[-wedges={int}] [-bands={int}] [-orient={bool}]" + NL +
"\t[-practice={bool}] [-base={int}] [-delay={int}]");
			return;
		}
		MballFrame frame = new MballFrame(MballInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle.requestFocusInWindow();
		frame.setLocation(MballFrame.xValue, MballFrame.yValue);
		frame.setVisible(true);
	}
}
